# -*- coding: utf-8 -*-


__all__ = ["atomstogrid", "mark_cavities", "cavity_triangles", "cavity_intersections"]


import numpy as np
import itertools
import sys
from math import ceil
from gr3 import triangulate


dimension = 3
dimensions = range(dimension)


def atomstogrid(
    grid,
    discrete_positions,
    radii_indices,
    discrete_radii,
    translation_vectors,
    discretization_grid,
):
    last_radius_index = -1  # (for reuse of sphere grids)
    atom_information = zip(
        range(len(discrete_positions)), radii_indices, discrete_positions
    )
    for atom_index, radius_index, real_discrete_position in atom_information:
        discrete_radius = discrete_radii[radius_index]
        cube_size = 2 * discrete_radius + 1
        if radius_index != last_radius_index:
            last_radius_index = radius_index
            cube_indices = (
                np.indices([cube_size] * 3).transpose(1, 2, 3, 0) - discrete_radius
            )
            sphere_grid = np.sum(cube_indices**2, axis=3) <= discrete_radius**2
            last_radius_index = radius_index
        for v in translation_vectors:
            discrete_position = [real_discrete_position[i] + v[i] for i in dimensions]
            for point in itertools.product(
                range(discrete_radius * 2 + 1), repeat=dimension
            ):
                if sphere_grid[point]:
                    p = [
                        point[i] - discrete_radius + discrete_position[i]
                        for i in dimensions
                    ]
                    if all([0 <= p[i] <= grid.shape[i] - 1 for i in dimensions]):
                        discretization_grid_value = discretization_grid[tuple(p)]
                        if discretization_grid_value == 0:
                            grid_value = grid[tuple(p)]
                            # check if the atom is the closest one
                            if grid_value == 0:
                                grid[tuple(p)] = atom_index + 1
                            else:
                                this_squared_distance = sum(
                                    [(x - y) ** 2 for x, y in zip(discrete_position, p)]
                                )
                                other_position = discrete_positions[grid_value - 1]
                                for v2 in translation_vectors:
                                    other_discrete_position = [
                                        other_position[i] + v2[i] for i in dimensions
                                    ]
                                    other_squared_distance = sum(
                                        [
                                            (x - y) ** 2
                                            for x, y in zip(other_discrete_position, p)
                                        ]
                                    )
                                    if other_squared_distance <= this_squared_distance:
                                        break
                                if this_squared_distance < other_squared_distance:
                                    grid[tuple(p)] = atom_index + 1


class Subgrid(object):
    def __init__(self, cubesize, grid_dimensions):
        self.cubesize = cubesize
        self.sgd = tuple(
            [2 + int(ceil(1.0 * d / cubesize)) + 2 for d in grid_dimensions]
        )
        self.sg = []
        for x in range(self.sgd[0]):
            self.sg.append([])
            for y in range(self.sgd[1]):
                self.sg[x].append([])
                for z in range(self.sgd[2]):
                    self.sg[x][y].append([[], [], []])  # SEG fault was here

    def add_atoms(self, atom_positions, translation_vectors):
        for atom_index, atom_position in enumerate(atom_positions):
            for v in translation_vectors:
                real_atom_position = [atom_position[i] + v[i] for i in dimensions]
                sgp = self.to_subgrid(real_atom_position)
                self.sg[sgp[0]][sgp[1]][sgp[2]][0].append(real_atom_position)

    def add_domains(self, domain_point_list, translation_vectors):
        for domain_index, domain_seed_points in enumerate(domain_point_list):
            for domain_seed_point in domain_seed_points:
                for v in translation_vectors:
                    real_domain_seed_point = [
                        domain_seed_point[i] + v[i] for i in dimensions
                    ]
                    sgp = self.to_subgrid(real_domain_seed_point)
                    self.sg[sgp[0]][sgp[1]][sgp[2]][1].append(real_domain_seed_point)
                    self.sg[sgp[0]][sgp[1]][sgp[2]][2].append(domain_index)

    def to_subgrid(self, position):
        sgp = [c // self.cubesize + 2 for c in position]
        for i in dimensions:
            if sgp[i] < 0:
                sgp[i] = 0
            if sgp[i] >= self.sgd[i]:
                sgp[i] = self.sgd[i] - 1
        return tuple(sgp)


def mark_cavities(
    domain_grid,
    discretization_grid,
    grid_dimensions,
    sg_cube_size,
    atom_positions,
    translation_vectors,
    domain_point_list,
    use_surface_points,
):

    # steps 1 to 3
    sg = Subgrid(sg_cube_size, grid_dimensions)
    sg.add_atoms(atom_positions, translation_vectors)
    sg.add_domains(domain_point_list, translation_vectors)

    # step 4
    grid = np.zeros(grid_dimensions, dtype=np.int64)
    for p in itertools.product(*map(range, grid_dimensions)):
        if use_surface_points:
            grid_value = domain_grid[p]
            if grid_value == 0:  # outside the volume
                grid[p] = 0
                possibly_in_cavity = False
            elif (
                grid_value < 0
            ):  # cavity domain (stored as: -index-1), therefore guaranteed to be in a cavity
                grid[p] = grid_value
                possibly_in_cavity = False  # is already marked
            elif (
                grid_value > 0
            ):  # in radius of atom (stored as: index+1), therefore possibly in a cavity
                grid[p] = 0
                possibly_in_cavity = True
        else:
            possibly_in_cavity = discretization_grid[p] == 0
        if possibly_in_cavity:
            # step 5
            min_squared_atom_distance = sys.maxsize
            sgp = sg.to_subgrid(p)
            for i in itertools.product((-1, 0, 1), repeat=dimension):
                sgci = [sgp[j] + i[j] for j in dimensions]
                for atom_position in sg.sg[sgci[0]][sgci[1]][sgci[2]][0]:
                    squared_atom_distance = sum(
                        [
                            (atom_position[j] - p[j]) * (atom_position[j] - p[j])
                            for j in dimensions
                        ]
                    )
                    min_squared_atom_distance = min(
                        min_squared_atom_distance, squared_atom_distance
                    )
            for i in itertools.product((-1, 0, 1), repeat=dimension):
                next = False
                sgci = [sgp[j] + i[j] for j in dimensions]
                for domain_index, domain_seed_point in zip(
                    sg.sg[sgci[0]][sgci[1]][sgci[2]][2],
                    sg.sg[sgci[0]][sgci[1]][sgci[2]][1],
                ):
                    squared_domain_seed_point_distance = sum(
                        [
                            (domain_seed_point[j] - p[j])
                            * (domain_seed_point[j] - p[j])
                            for j in dimensions
                        ]
                    )
                    if squared_domain_seed_point_distance < min_squared_atom_distance:
                        grid[p] = -domain_index - 1
                        next = True
                        break
                if next:
                    break
    return grid


def cavity_triangles(
    cavity_grid, cavity_indices, isolevel, step, offset, discretization_grid
):
    cavity_triangles = []
    cavity_surface_areas = []
    grid = np.zeros(cavity_grid.shape, dtype=bool)

    for cavity_index in cavity_indices:
        grid = np.logical_or(grid, cavity_grid == -(cavity_index + 1))
    views = []
    for x, y, z in itertools.product(*map(range, (3, 3, 3))):
        view = grid[
            x : grid.shape[0] - 2 + x,
            y : grid.shape[1] - 2 + y,
            z : grid.shape[2] - 2 + z,
        ]
        views.append(view)
    grid = np.zeros(grid.shape, np.uint16)
    grid[:, :, :] = 0
    grid[1:-1, 1:-1, 1:-1] = sum(views) + 100

    vertices, normals = triangulate(grid, (1, 1, 1), (0, 0, 0), 100 + isolevel)
    discrete_vertices = np.array(np.floor(vertices + 0.5), dtype=np.int32)
    vertices *= np.tile(step, (vertices.shape[0], 3, 1))
    vertices += np.tile(offset, (vertices.shape[0], 3, 1))
    normals /= np.tile(step, (normals.shape[0], 3, 1))

    cavity_surface_area = 0
    for cavity_triangle, discrete_triangle in zip(vertices, discrete_vertices):
        any_outside = False
        for vertex, discrete_vertex in zip(cavity_triangle, discrete_triangle):
            if discretization_grid[tuple(discrete_vertex)] != 0:
                any_outside = True
                break
        if not any_outside:
            v1, v2, v3 = cavity_triangle
            a = v2 - v1
            b = v3 - v1
            triangle_surface_area = np.linalg.norm(np.cross(a, b)) * 0.5
            cavity_surface_area += triangle_surface_area

    return vertices, normals, cavity_surface_area


def cavity_intersections(grid, num_domains):
    intersection_table = np.zeros((num_domains, num_domains), dtype=np.int8)
    directions = []
    for dx, dy, dz in itertools.product((0, 1), repeat=3):
        if any((dx > 0, dy > 0, dz > 0)):
            directions.append((dx, dy, dz))
    for p in itertools.product(*[range(x - 1) for x in grid.shape]):
        domain1 = -grid[p] - 1
        if domain1 != -1:
            for direction in directions:
                p2 = tuple([p[i] + direction[i] for i in dimensions])
                domain2 = -grid[p2] - 1
                if domain2 != -1:
                    intersection_table[domain1][domain2] = 1
                    intersection_table[domain2][domain1] = 1
    return intersection_table


def mark_translation_vectors(grid, translation_vectors):
    # step 5
    for p in itertools.product(*map(range, grid.shape)):
        equivalent_points = [
            [p[i] + v[i] for i in dimensions] for v in translation_vectors
        ]
        valid_equivalent_points = [
            tuple(point)
            for point in equivalent_points
            if all([0 <= point[i] <= grid.shape[i] - 1 for i in dimensions])
        ]
        if grid[p] == 0:
            equivalent_points_inside = [
                point for point in valid_equivalent_points if grid[point] == 0
            ]
            for point in equivalent_points_inside:
                grid[point] = 1
    # step 6 & 7
    for p in itertools.product(*map(range, grid.shape)):
        equivalent_points = [
            ([p[i] + v[i] for i in dimensions], vi)
            for vi, v in enumerate(translation_vectors)
        ]
        valid_equivalent_points = [
            (tuple(point), vi)
            for point, vi in equivalent_points
            if all([0 <= point[i] <= grid.shape[i] - 1 for i in dimensions])
        ]
        if grid[p] == 1:
            equivalent_points_inside = [
                (point, vi) for point, vi in valid_equivalent_points if grid[point] == 0
            ]
            if not equivalent_points_inside:
                nearest_to_center = p
                nearest_to_center_index = -1  # -1 -> -(-1+1) == 0
                min_d_center = sum(
                    [
                        (p[i] - grid.shape[i] / 2) * (p[i] - grid.shape[i] / 2)
                        for i in dimensions
                    ]
                )
                for p2, vi in valid_equivalent_points:
                    d_center = sum(
                        [
                            (p2[i] - grid.shape[i] / 2) * (p2[i] - grid.shape[i] / 2)
                            for i in dimensions
                        ]
                    )
                    if d_center < min_d_center:
                        min_d_center = d_center
                        nearest_to_center = p2
                        nearest_to_center_index = vi
                grid[nearest_to_center] = 0
                grid[p] = -(nearest_to_center_index + 1)
            else:
                translation_vector_index = equivalent_points_inside[0][1]
                grid[p] = -(translation_vector_index + 1)



# coding: utf-8

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import math
import numpy as np
import numpy.linalg as la


def calculate_gyration_tensor_parameters(points):
    """
    Calculates the gyration tensor parameters R_g^2, η, c, κ from a list of
    all points inside a cavity.
     - R_g^2 is the squared gyration radius
     - η is the asphericity
     - c is the acylindricity
     - κ is the anisotropy
    """

    points = np.array(points, dtype=np.float64)
    mean = np.mean(points, axis=0)
    points -= mean

    gyration_tensor = np.zeros((3, 3))
    for i in range(3):
        for j in range(i, 3):
            gyration_tensor[i, j] = np.dot(points[:, i], points[:, j])
            gyration_tensor[j, i] = gyration_tensor[i, j]
    # cell volume is constant, cavity volume is proportional to len(points)
    gyration_tensor /= len(points)

    eigvals = list(sorted(la.eigvalsh(gyration_tensor), reverse=True))

    squared_gyration_radius = sum(eigvals)
    if squared_gyration_radius > 0:
        asphericity = (
            eigvals[0] - 0.5 * (eigvals[1] + eigvals[2])
        ) / squared_gyration_radius
        acylindricity = (eigvals[1] - eigvals[2]) / squared_gyration_radius
        anisotropy = (asphericity**2 + 0.75 * acylindricity**2) ** 0.5
    else:
        asphericity = 0
        acylindricity = 0
        anisotropy = 0
    return mean, squared_gyration_radius, asphericity, acylindricity, anisotropy


# Test code:


def generate_box_points(offset, side_length, n):
    return generate_cuboid_points(offset, (side_length, side_length, side_length), n)


def generate_cuboid_points(offset, side_lengths, n):
    offset = np.array(offset)
    interval = 0.5 * max(side_lengths) * np.linspace(-1, 1, n)
    points = []
    for x in interval:
        if abs(x) > 0.5 * side_lengths[0]:
            continue
        for y in interval:
            if abs(y) > 0.5 * side_lengths[1]:
                continue
            for z in interval:
                if abs(z) > 0.5 * side_lengths[2]:
                    continue
                points.append((x, y, z) + offset)
    return points


def generate_sphere_points(offset, radius, n):
    offset = np.array(offset)
    interval = radius * np.linspace(-1, 1, n)
    points = []
    for x in interval:
        for y in interval:
            for z in interval:
                if la.norm((x, y, z)) <= radius:
                    points.append((x, y, z) + offset)
    return points


def generate_cylinder_points(offset, radius, length, n):
    offset = np.array(offset)
    interval = max(radius, length / 2) * np.linspace(-1, 1, n)
    points = []
    for x in interval:
        for y in interval:
            for z in interval:
                if abs(z) < length / 2 and la.norm((x, y)) <= radius:
                    points.append((x, y, z) + offset)
    return points


def main():
    silly_offset = (-2, 17.3, 42)
    print(
        "box      (a=1):            ",
        calculate_gyration_tensor_parameters(generate_box_points(silly_offset, 1, 100)),
    )
    print(
        "box      (a=2):            ",
        calculate_gyration_tensor_parameters(generate_box_points(silly_offset, 2, 100)),
    )
    print(
        "cuboid   (a=1, b=2, c=1):  ",
        calculate_gyration_tensor_parameters(
            generate_cuboid_points(silly_offset, (1, 2, 1), 100)
        ),
    )
    print(
        "cuboid   (a=1, b=20, c=1): ",
        calculate_gyration_tensor_parameters(
            generate_cuboid_points(silly_offset, (1, 20, 1), 100)
        ),
    )
    print(
        "sphere   (r=1):            ",
        calculate_gyration_tensor_parameters(
            generate_sphere_points(silly_offset, 1, 100)
        ),
    )
    print(
        "sphere   (r=2):            ",
        calculate_gyration_tensor_parameters(
            generate_sphere_points(silly_offset, 2, 100)
        ),
    )
    print(
        "cylinder (r=1, l=1):       ",
        calculate_gyration_tensor_parameters(
            generate_cylinder_points(silly_offset, 1, 1, 100)
        ),
    )
    print(
        "cylinder (r=1, l=20):      ",
        calculate_gyration_tensor_parameters(
            generate_cylinder_points(silly_offset, 1, 20, 100)
        ),
    )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-


from __future__ import absolute_import


from PySide6 import QtCore, QtWidgets
from ...config.cutoff_history import cutoff_history
from ...config.cutoff_history import HistoryEntry
from ..cutoff_history_table import CutoffHistoryTable


class CutoffHistoryDialog(QtWidgets.QDialog):
    def __init__(self, parent, elements, preferred_filenames_with_frames=None):
        super().__init__(parent)
        history = cutoff_history.filtered_history(
            elements, preferred_filenames_with_frames=preferred_filenames_with_frames
        )
        self._init_ui(history)

    def _init_ui(self, history):
        self.gb_history = QtWidgets.QGroupBox("History")
        self.tw_cutoff = CutoffHistoryTable(history)
        self.pb_ok = QtWidgets.QPushButton("Ok")
        self.pb_ok.setEnabled(False)
        self.pb_cancel = QtWidgets.QPushButton("Cancel")

        self.la_main = QtWidgets.QVBoxLayout()
        self.la_history = QtWidgets.QVBoxLayout()
        self.la_ok_cancel = QtWidgets.QHBoxLayout()
        self.la_main.addWidget(self.gb_history)
        self.la_history.addWidget(self.tw_cutoff)
        self.la_ok_cancel.addStretch()
        self.la_ok_cancel.addWidget(self.pb_ok)
        self.la_ok_cancel.addStretch()
        self.la_ok_cancel.addWidget(self.pb_cancel)
        self.la_ok_cancel.addStretch()
        self.la_ok_cancel.setContentsMargins(0, 0, 0, 0)
        self.la_main.addLayout(self.la_ok_cancel)
        self.la_main.setSizeConstraint(QtWidgets.QLayout.SetFixedSize)
        self.gb_history.setLayout(self.la_history)
        self.setLayout(self.la_main)

        self.tw_cutoff.itemSelectionChanged.connect(
            lambda *args: self.pb_ok.setEnabled(
                len(self.tw_cutoff.selectedIndexes()) > 0
            )
        )
        self.pb_ok.clicked.connect(lambda event: self.accept())
        self.pb_cancel.clicked.connect(lambda event: self.reject())

    def accept(self):
        cutoff_history.remove_list(self.tw_cutoff.history_entries_for_deletion)
        cutoff_history.save()
        return super(CutoffHistoryDialog, self).accept()

    @property
    def history(self):
        return self.tw_cutoff.history

    @property
    def selected_radii(self):
        return self.history[self.tw_cutoff.selectedIndexes()[0].row()].radii



from . import algorithm
from .algorithm import ObjectType
from .util.graph import GraphForSplitAndMerge


def start_split_and_merge_pipeline(
    data, mask, atoms, combined_translation_vectors, get_translation_vector, object_type
):
    def is_relevant_part(hom_image_data_part):
        return algorithm.is_relevant_part(hom_image_data_part, object_type)

    graph = GraphForSplitAndMerge(data, mask, get_translation_vector, is_relevant_part)
    algorithm.split(data, mask, graph, object_type)
    algorithm.merge(data, graph)
    algorithm.add_periodic_neighbors(graph)
    algorithm.merge_periodic_border(data, graph)
    areas, non_translated_areas, cyclic_area_indices = graph.get_all_areas(
        apply_translation=True, with_non_translated_nodes=True, mark_cyclic_areas=True
    )
    areas_cells, non_translated_areas_cells = map(
        algorithm.get_domain_area_cells, [areas, non_translated_areas]
    )
    if object_type == ObjectType.DOMAIN:
        algorithm.mark_domain_points(data, non_translated_areas)
        centers = algorithm.calculate_domain_centers(
            atoms, combined_translation_vectors, non_translated_areas
        )
        surface_cells = algorithm.get_domain_surface_cells(
            data, mask, non_translated_areas
        )

        return (
            centers,
            areas_cells,
            non_translated_areas_cells,
            surface_cells,
            cyclic_area_indices,
        )
    else:
        return areas_cells, non_translated_areas_cells, cyclic_area_indices



from PySide6 import QtCore, QtGui, QtWidgets


class AboutDialog(QtWidgets.QDialog):
    """
    Einfacher About-Dialog, der den Titel, das Icon, eine Beschreibung und eine Liste von Autoren anzeigt.
    Autoren werden als Liste von Tupeln uebergeben, in dem Format (Name (str), Mail (str))
    """

    class MailLabel(QtWidgets.QLabel):
        def __init__(self, mail, parent):
            super().__init__('<a href="mailto:%s">%s</a>' % (mail, mail), parent)
            self.mail = mail

        def mousePressEvent(self, event):
            url = QtCore.QUrl("mailto:%s" % self.mail)
            QtGui.QDesktopServices.openUrl(url)

    def __init__(self, parent, description, authors, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)

        self.title = self.parent().window().windowTitle()
        self.icon = self.parent().window().windowIcon()
        self.description = description
        self.authors = authors

        self._init_ui()

    def _init_ui(self):
        # Label
        self.lb_title = QtWidgets.QLabel("<h2>%s</h2>" % self.title, self)
        self.lb_description = QtWidgets.QLabel("<p>%s</p>" % self.description, self)
        self.lb_author = QtWidgets.QLabel("<b>%s</b>" % "Autoren:", self)
        self.lb_authors = []
        for author, mail in self.authors:
            lb_author = QtWidgets.QLabel(author, self)
            lb_mail = self.MailLabel(mail, self)
            self.lb_authors.append((lb_author, lb_mail))

        # Icon
        self.pm_icon = QtWidgets.QLabel(self)
        self.pm_icon.setPixmap(self.icon.pixmap(128, 128))

        # Layout
        self.la_main_layout = QtWidgets.QVBoxLayout()
        self.la_icon_desc_layout = QtWidgets.QHBoxLayout()
        self.la_desc_author_layout = QtWidgets.QVBoxLayout()
        self.la_authors = QtWidgets.QGridLayout()

        self.la_main_layout.addWidget(self.lb_title, alignment=QtCore.Qt.AlignHCenter)
        self.la_main_layout.addLayout(self.la_icon_desc_layout, stretch=1)
        self.la_icon_desc_layout.addWidget(self.pm_icon)
        self.la_icon_desc_layout.addLayout(self.la_desc_author_layout, stretch=1)
        self.la_desc_author_layout.addWidget(self.lb_description)
        self.la_desc_author_layout.addSpacing(5)
        self.la_desc_author_layout.addWidget(self.lb_author)
        self.la_desc_author_layout.addLayout(self.la_authors)
        self.la_authors.setContentsMargins(5, 0, 0, 0)
        for i, (lb_author, lb_mail) in enumerate(self.lb_authors):
            self.la_authors.addWidget(lb_author, i, 0)
            self.la_authors.addWidget(lb_mail, i, 1)

        self.la_main_layout.setSizeConstraint(QtWidgets.QLayout.SetFixedSize)
        self.setLayout(self.la_main_layout)



from . import configobj
from . import validate
import os
import os.path
import inspect

CONFIG_DIRECTORY = "~/.pymoldyn/"  # MUST be written with ~ to save a path in the config file that is relative to the user's home directory
CONFIG_FILE = os.path.expanduser("%s/config.cfg" % CONFIG_DIRECTORY)
CONFIG_SPEC_FILE = os.path.expanduser("%s/config.spec" % CONFIG_DIRECTORY)

# second string is the list type name
type_dict = {
    int: ("integer", "int"),
    float: ("float", "float"),
    str: ("string", "string"),
    str: ("string", "string"),
    bool: ("boolean", "bool"),
}


class ConfigNode(object):

    def __init__(self):
        pass


class Configuration(ConfigNode):
    """
    Configuration Object that contains the application settings
    """

    class Colors(ConfigNode):

        def __init__(self):
            self.surface_cavity = [0.2, 0.4, 1.0]
            self.domain = [0.0, 1.0, 0.5]
            self.center_cavity = [0.9, 0.4, 0.2]
            self.background = [0.0, 0.0, 0.0]
            self.bounding_box = [1.0, 1.0, 1.0]
            self.bonds = [0.8, 0.8, 0.8]

    class OpenGL(ConfigNode):

        def __init__(self):
            # camera_position =
            # offset          = (0.0, 0.0, 0.0)
            self.gl_window_size = [1200, 400]
            self.atom_radius = 0.4
            self.bond_radius = 0.1
            pass

    class Computation(ConfigNode):
        def __init__(self):
            self.std_cutoff_radius = 2.8
            self.std_resolution = 64
            self.max_cachefiles = 0

    class Path(ConfigNode):

        def __init__(self):
            self.cache_dir = os.path.join(CONFIG_DIRECTORY, "cache")
            self.ffmpeg = "/usr/local/bin/ffmpeg"
            self.result_dir = os.path.join(CONFIG_DIRECTORY, "results")

    def __init__(self):
        # standard configuration
        self.Colors = Configuration.Colors()
        self.OpenGL = Configuration.OpenGL()
        self.Computation = Configuration.Computation()
        self.Path = Configuration.Path()

        self.window_position = [-1, -1]
        self.recent_files = [""]
        self.max_files = 5

        self._file = ConfigFile(self)

    def add_recent_file(self, filename):
        if len(self.recent_files) == 1 and not self.recent_files[0]:
            self.recent_files[0] = filename
        elif len(self.recent_files) == self.max_files:
            self.recent_files.pop(-1)
            self.recent_files.insert(0, filename)
        else:
            self.recent_files.insert(0, filename)
        self.save()

    def save(self):
        """
        write configuration to file
        """

        self._file.save()

    def read(self):
        """
        read configuration from file
        """
        self._file = ConfigFile(self)
        self._file.read()


class ConfigFile(object):
    """
    ConfigFile that parses the settings to a configuration file using ConfigObj 4
    """

    def __init__(self, cfg):
        self.config = cfg

    @staticmethod
    def _create_needed_parent_directories(filename):
        dirname = os.path.dirname(filename)
        if not os.path.exists(dirname):
            os.makedirs(dirname)

    def generate_configspec(self):
        """
        generates the type specification for the configuration data
        """
        spec_file = configobj.ConfigObj(CONFIG_SPEC_FILE)
        self.generate_spec_for_section(self.file, spec_file)
        # TODO: better error handling
        try:
            self._create_needed_parent_directories(CONFIG_SPEC_FILE)
            spec_file.write()
        except IOError as e:
            print("IOError in ConfigFile.generate_configspec")

    def generate_spec_for_section(self, section, spec_section):
        """
        recursive type specification for each subtree
        """
        for scalar in section.scalars:
            t = type(section[scalar])
            type_string = (
                type_dict[t][0]
                if t is not list
                else type_dict[type(section[scalar][0])][1] + "_list"
            )
            spec_section[scalar] = type_string
        for sect in section.sections:
            spec_section[sect] = {}
            self.generate_spec_for_section(section[sect], spec_section[sect])

    def save(self):
        """
        recursively reads the object and saves it to the ConfigFile object and finally writes it into the file
        """
        self.file = configobj.ConfigObj(CONFIG_FILE)
        self.parse_node_to_section(self.config, self.file)

        # TODO: better error handling
        try:
            self._create_needed_parent_directories(CONFIG_FILE)
            self.file.write()
            self.generate_configspec()
            self.file.write()
        except IOError as e:
            print("IOError in ConfigFile.save")

    def parse_node_to_section(self, node, section):
        """
        parses a ConfigNode to file object
        """
        for attr_str in dir(node):
            attr = getattr(node, attr_str)
            if isinstance(attr, ConfigNode):
                section[type(attr).__name__] = {}
                self.parse_node_to_section(attr, section[type(attr).__name__])
            elif not inspect.ismethod(attr) and not attr_str.startswith("_"):
                section[attr_str] = attr
            else:
                pass
                # print attr_str, 'NOT PROCESSED'

    def read(self):
        """
        read a configuration from file
        """
        if not os.path.isfile(CONFIG_SPEC_FILE) or not os.path.isfile(CONFIG_FILE):
            self.save()
        else:
            validator = validate.Validator()
            self.file = configobj.ConfigObj(CONFIG_FILE, configspec=CONFIG_SPEC_FILE)
            self.file.validate(validator)
            self.parse_section_to_node(self.file, self.config)

    def parse_section_to_node(self, section, node):
        """
        parses a config section to config object
        """
        for scalar in section.scalars:
            setattr(node, scalar, section[scalar])
        for sec in section.sections:
            self.parse_section_to_node(section[sec], getattr(node, sec))


config = Configuration()
config.read()



#  import numpy as np
#
#  def calculate_domain_centers(atoms, combined_translation_vectors, domains):
#  atom_pos = np.array(atoms, dtype=np.int32)
#  combined_trans_vecs = np.vstack(([0, 0, 0], combined_translation_vectors))
#  centers_list = []
#
#  for domain_nodes in domains:
#  max_distance = -1
#  for node_pos, node_dim in domain_nodes:
#  node_pos = np.array(node_pos)
#  node_dim = np.array(node_dim)
#  for x in range(node_pos[0], node_pos[0] + node_dim[0]):
#  for y in range(node_pos[1], node_pos[1] + node_dim[1]):
#  for z in range(node_pos[2], node_pos[2] + node_dim[2]):
#  min_distance = np.iinfo(np.int32).max
#  for atom in atom_pos:
#  for trans_vec in combined_trans_vecs:
#  diff = np.array([x, y, z]) - atom + trans_vec
#  current_distance = np.sum(diff ** 2)
#  if current_distance < min_distance:
#  min_distance = current_distance
#  if min_distance > max_distance:
#  max_distance = min_distance
#  current_center = np.array([x, y, z])
#  centers_list.append(tuple(current_center))
#  return centers_list
#



from PySide6 import QtCore, QtGui, QtWidgets
import sys
from collections import OrderedDict
import functools
from ...config.configuration import config, Configuration
from ..gl_widget import UpdateGLEvent, GLWidget
from ..cutoff_history_table import CutoffHistoryTable
from ..cutoff_preset_table import CutoffPresetTable
from ...config.cutoff_history import cutoff_history
from ...config.cutoff_presets import cutoff_presets


class GraphicsSettingsPage(QtWidgets.QWidget):

    def __init__(self, parent, cfg):
        QtWidgets.QWidget.__init__(self, parent)

        self._config = cfg
        self.values = OrderedDict(
            (
                ("atom_radius", "atom radius"),
                ("bond_radius", "bond radius"),
            )
        )

        self.lineedit_dict = {}
        grid = QtWidgets.QGridLayout()
        for i, (attr_str, label) in enumerate(self.values.items()):
            cfg_comp = getattr(self._config, "OpenGL")

            l = QtWidgets.QLabel(label, self)
            t = QtWidgets.QLineEdit(self)
            t.setText(str(getattr(cfg_comp, attr_str)))
            self.lineedit_dict[attr_str] = t
            t.textChanged.connect(self.any_change)
            grid.addWidget(l, i, 0)
            grid.addWidget(t, i, 1)

        self.colors = OrderedDict(
            (
                ("background", "Background"),
                ("bounding_box", "Bounding Box"),
                ("bonds", "Bonds"),
                ("domain", "Cavity (domain)"),
                ("surface_cavity", "Cavity (surface method)"),
                ("center_cavity", "Cavity (Center method)"),
            )
        )

        self.button_dict = {}
        layout = QtWidgets.QGridLayout()
        layout.setSpacing(5)
        for index, (attr_str, btn_str) in enumerate(self.colors.items()):

            pix = QtGui.QPixmap(50, 50)
            cfg_clr = getattr(self._config, "Colors")
            tmp = [int(f * 255) for f in getattr(cfg_clr, attr_str)]
            current_color = QtGui.QColor(*tmp)
            pix.fill(current_color)

            b = QtWidgets.QPushButton(None, self)
            b.setFixedSize(50, 50)
            b.setFocusPolicy(QtCore.Qt.StrongFocus)
            self.button_dict[attr_str] = b
            b.clicked.connect(
                lambda _, arg1=attr_str, arg2=current_color: self.show_color_dialog(
                    arg1, arg2
                )
            )
            b.setIcon(QtGui.QIcon(pix))

            layout.addWidget(b, index, 0)
            layout.addWidget(QtWidgets.QLabel(btn_str, self), index, 1)

        box = QtWidgets.QVBoxLayout()
        box.addLayout(grid)
        box.addStretch()
        box.addLayout(layout)
        self.setLayout(box)
        self.setLayout(box)
        self.show()

    def keyPressEvent(self, event):
        pass

    def any_change(self):
        cfg_comp = getattr(self._config, "OpenGL")
        for i, (attr_str, label) in enumerate(self.values.items()):
            try:
                setattr(cfg_comp, attr_str, float(self.lineedit_dict[attr_str].text()))
            except ValueError:
                setattr(cfg_comp, attr_str, 0.0)

    def show_color_dialog(self, s, previous_color):
        color = QtWidgets.QColorDialog.getColor(initial=previous_color)
        if color.isValid():
            pix = QtGui.QPixmap(50, 50)
            pix.fill(color)
            self.button_dict[s].setIcon(QtGui.QIcon(pix))
            for i, clr_val in enumerate(color.getRgb()[:3]):
                tmp_cfg_clr = getattr(self._config, "Colors")
                getattr(tmp_cfg_clr, s)[i] = clr_val / 255.0


class ComputationSettingsPage(QtWidgets.QWidget):

    def __init__(self, parent, cfg):
        QtWidgets.QWidget.__init__(self, parent)
        self._config = cfg
        self.values = OrderedDict((("std_cutoff_radius", "default cutoff radius"),))

        self.lineedit_dict = {}
        box = QtWidgets.QVBoxLayout()
        grid = QtWidgets.QGridLayout()
        for i, (attr_str, label) in enumerate(self.values.items()):
            cfg_comp = getattr(self._config, "Computation")

            l = QtWidgets.QLabel(label, self)
            t = QtWidgets.QLineEdit(self)
            t.setText(str(getattr(cfg_comp, attr_str)))
            self.lineedit_dict[attr_str] = t

            # self.connect(t, QtCore.SIGNAL("editingFinished()"), lambda who=attr_str: self.value_edited(who))
            t.textChanged.connect(self.any_change)
            grid.addWidget(l, i, 0)
            grid.addWidget(t, i, 1)

        self.tw_cutoff_history = CutoffHistoryTable(cutoff_history.history)
        pb_clear_cutoff_history = QtWidgets.QPushButton(
            self.style().standardIcon(QtWidgets.QStyle.SP_TrashIcon),
            "Clear cutoff history",
        )
        pb_clear_cutoff_history.clicked.connect(self.clear_cutoff_history_requested)
        if len(cutoff_history.history) == 0:
            self.tw_cutoff_history.setVisible(False)
            pb_clear_cutoff_history.setVisible(False)

        self.tw_cutoff_presets = CutoffPresetTable(cutoff_presets.presets)
        pb_clear_cutoff_presets = QtWidgets.QPushButton(
            self.style().standardIcon(QtWidgets.QStyle.SP_TrashIcon),
            "Clear cutoff presets",
        )
        pb_clear_cutoff_presets.clicked.connect(self.clear_cutoff_presets_requested)
        if len(cutoff_presets.presets) == 0:
            self.tw_cutoff_presets.setVisible(False)
            pb_clear_cutoff_presets.setVisible(False)

        box.addLayout(grid)
        box.addWidget(self.tw_cutoff_history, 0, QtCore.Qt.AlignHCenter)
        box.addWidget(pb_clear_cutoff_history, 0, QtCore.Qt.AlignRight)
        box.addWidget(self.tw_cutoff_presets, 0, QtCore.Qt.AlignHCenter)
        box.addWidget(pb_clear_cutoff_presets, 0, QtCore.Qt.AlignRight)
        box.addStretch()
        self.setLayout(box)
        self.show()

    def clear_cutoff_history_requested(self):
        message = QtWidgets.QMessageBox()
        message.setStandardButtons(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
        message.setText(
            "Do you really want to clear the cutoff radii history for all previous done calculations?\n\n"
            + "(The history data will be cleared when you hit the ok button of the settings dialog.)"
        )
        answer = message.exec_()
        if answer == QtWidgets.QMessageBox.Yes:
            self.tw_cutoff_history.clear_entries()

    def clear_cutoff_presets_requested(self):
        message = QtWidgets.QMessageBox()
        message.setStandardButtons(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
        message.setText(
            "Do you really want to delete all cutoff presets?\n\n"
            + "(The presets will be removed irrevocably when you hit the ok button of the settings dialog.)"
        )
        answer = message.exec_()
        if answer == QtWidgets.QMessageBox.Yes:
            self.tw_cutoff_presets.clear_entries()

    def any_change(self):
        cfg_comp = getattr(self._config, "Computation")
        for i, (attr_str, label) in enumerate(self.values.items()):
            setattr(cfg_comp, attr_str, float(self.lineedit_dict[attr_str].text()))

    # def value_edited(self, s):
    #     cfg_comp = getattr(self._config, 'Computation')
    #     setattr(cfg_comp, s, float(self.lineedit_dict[s].text()))
    #     print 'VALUE EDITED'


class PathSettingsPage(QtWidgets.QWidget):

    def __init__(self, parent, cfg):
        QtWidgets.QWidget.__init__(self, parent)

        self._config = cfg
        self.path = OrderedDict(
            (
                ("cache_dir", "Cache directory"),
                ("result_dir", "Result directory"),
                ("ffmpeg", "ffmpeg path"),
            )
        )

        self.lineedit_dict = {}
        box = QtWidgets.QVBoxLayout()
        grid = QtWidgets.QGridLayout()
        for i, (attr_str, label) in enumerate(self.path.items()):
            cfg_path = getattr(self._config, "Path")

            l = QtWidgets.QLabel(label, self)
            t = QtWidgets.QLineEdit(self)
            t.setText(getattr(cfg_path, attr_str))
            self.lineedit_dict[attr_str] = t

            t.editingFinished.connect(lambda who=attr_str: self.path_edited(who))
            grid.addWidget(l, i, 0)
            grid.addWidget(t, i, 1)
        box.addLayout(grid)
        box.addStretch()
        self.setLayout(box)
        self.show()

    def path_edited(self, s):
        cfg_path = getattr(self._config, "Path")
        setattr(cfg_path, s, str(self.lineedit_dict[s].text()))


class SettingsDialog(QtWidgets.QDialog):

    def __init__(self, parent=None):
        QtWidgets.QDialog.__init__(self, parent)

        # temporary Configuration which saves the current changes
        self._tmp = Configuration()
        self._tmp.read()

        vbox = QtWidgets.QVBoxLayout()
        tab_widget = QtWidgets.QTabWidget()
        graphics_page = GraphicsSettingsPage(tab_widget, self._tmp)
        path_page = PathSettingsPage(tab_widget, self._tmp)
        comp_page = ComputationSettingsPage(tab_widget, self._tmp)
        self.cutoff_history_entries_for_deletion = (
            lambda: comp_page.tw_cutoff_history.history_entries_for_deletion
        )
        self.cutoff_preset_entries_for_deletion = (
            lambda: comp_page.tw_cutoff_presets.preset_entries_for_deletion
        )

        # Ok, Cancel and Restore defaults Buttons
        ok = QtWidgets.QPushButton("Ok", self)
        cancel = QtWidgets.QPushButton("Cancel", self)
        restore_btn = QtWidgets.QPushButton("Restore defaults", self)

        ok.clicked.connect(self.ok)
        cancel.clicked.connect(self.cancel)
        restore_btn.clicked.connect(self.restore_defaults)

        tab_widget.addTab(graphics_page, "graphics")
        tab_widget.addTab(path_page, "path")
        tab_widget.addTab(comp_page, "computation")

        vbox.addWidget(tab_widget)

        hbox = QtWidgets.QHBoxLayout()
        hbox.addStretch()
        hbox.addWidget(ok)
        hbox.addStretch()
        hbox.addWidget(cancel)
        hbox.addStretch()
        hbox.addWidget(restore_btn)
        hbox.addStretch()

        vbox.addLayout(hbox)
        self.setLayout(vbox)

        self.exec_()

    def ok(self):
        global config
        self._tmp.save()
        config.read()
        cutoff_history.remove_list(self.cutoff_history_entries_for_deletion())
        cutoff_history.save()
        cutoff_presets.remove_list(self.cutoff_preset_entries_for_deletion())
        cutoff_presets.save()
        self.accept()
        for widget in QtWidgets.QApplication.topLevelWidgets():
            for gl_widget in widget.findChildren(GLWidget):
                gl_widget.update_needed = True
                QtWidgets.QApplication.postEvent(gl_widget, UpdateGLEvent())

    def restore_defaults(self):
        reply = QtWidgets.QMessageBox.question(
            self,
            "Message",
            "Are you sure to restore the defaults?",
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
            QtWidgets.QMessageBox.No,
        )
        if reply == QtWidgets.QMessageBox.Yes:
            self._tmp = Configuration()
            self._tmp.save()
            config.read()
            self.accept()

    def cancel(self):
        self.reject()


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    config.read()
    dia = SettingsDialog()
    sys.exit(app.exec_())



import itertools as it


def iterate_node_border(node, func):
    node_x, node_y, node_z = node[0]
    node_w, node_h, node_d = node[1]

    border_points = [
        (x, node_y - 1, z)
        for x in range(node_x, node_x + node_w + 1)
        for z in range(node_z, node_z + node_d + 1)
    ]
    border_points.extend(
        [
            (x, node_y + node_h, z)
            for x in range(node_x, node_x + node_w + 1)
            for z in range(node_z, node_z + node_d + 1)
        ]
    )
    border_points.extend(
        [
            (node_x - 1, y, z)
            for y in range(node_y - 1, node_y + node_h + 1)
            for z in range(node_z - 1, node_z + node_d + 1)
        ]
    )
    border_points.extend(
        [
            (node_x + node_w, y, z)
            for y in range(node_y, node_y + node_h)
            for z in range(node_z, node_z + node_d + 1)
        ]
    )
    border_points.extend(
        [
            (x, y, node_z - 1)
            for x in range(node_x, node_x + node_w + 1)
            for y in range(node_y - 1, node_y + node_h + 1)
        ]
    )
    border_points.extend(
        [
            (x, y, node_z + node_d)
            for x in range(node_x, node_x + node_w)
            for y in range(node_y, node_y + node_h)
        ]
    )

    for border_x, border_y, border_z in border_points:
        func(border_x, border_y, border_z)


def iterate_node_border_with_adjacent_node_cells(node, func):
    node_x, node_y, node_z = node[0]
    node_w, node_h, node_d = node[1]

    border_points = [
        (
            x,
            node_y - 1,
            z,
            tuple(
                it.product(
                    range(max(x - 1, node_x), min(x + 2, node_x + node_w)),
                    [node_y],
                    range(max(z - 1, node_z), min(z + 2, node_z + node_d)),
                )
            ),
        )
        for x in range(node_x, node_x + node_w + 1)
        for z in range(node_z, node_z + node_d + 1)
    ]
    border_points.extend(
        [
            (
                x,
                node_y + node_h,
                z,
                tuple(
                    it.product(
                        range(max(x - 1, node_x), min(x + 2, node_x + node_w)),
                        [node_y + node_h - 1],
                        range(max(z - 1, node_z), min(z + 2, node_z + node_d)),
                    )
                ),
            )
            for x in range(node_x, node_x + node_w + 1)
            for z in range(node_z, node_z + node_d + 1)
        ]
    )
    border_points.extend(
        [
            (
                node_x - 1,
                y,
                z,
                tuple(
                    it.product(
                        [node_x],
                        range(max(y - 1, node_y), min(y + 2, node_y + node_h)),
                        range(max(z - 1, node_z), min(z + 2, node_z + node_d)),
                    )
                ),
            )
            for y in range(node_y - 1, node_y + node_h + 1)
            for z in range(node_z - 1, node_z + node_d + 1)
        ]
    )
    border_points.extend(
        [
            (
                node_x + node_w,
                y,
                z,
                tuple(
                    it.product(
                        [node_x + node_w - 1],
                        range(max(y - 1, node_y), min(y + 2, node_y + node_h)),
                        range(max(z - 1, node_z), min(z + 2, node_z + node_d)),
                    )
                ),
            )
            for y in range(node_y, node_y + node_h)
            for z in range(node_z, node_z + node_d + 1)
        ]
    )
    border_points.extend(
        [
            (
                x,
                y,
                node_z - 1,
                tuple(
                    it.product(
                        range(max(x - 1, node_x), min(x + 2, node_x + node_w)),
                        range(max(y - 1, node_y), min(y + 2, node_y + node_h)),
                        [node_z],
                    )
                ),
            )
            for x in range(node_x, node_x + node_w + 1)
            for y in range(node_y - 1, node_y + node_h + 1)
        ]
    )
    border_points.extend(
        [
            (
                x,
                y,
                node_z + node_d,
                tuple(
                    it.product(
                        range(max(x - 1, node_x), min(x + 2, node_x + node_w)),
                        range(max(y - 1, node_y), min(y + 2, node_y + node_h)),
                        [node_z + node_d - 1],
                    )
                ),
            )
            for x in range(node_x, node_x + node_w)
            for y in range(node_y, node_y + node_h)
        ]
    )

    for border_x, border_y, border_z, adjacent_node_cells in border_points:
        func(border_x, border_y, border_z, adjacent_node_cells)



# -*- coding: utf-8 -*-


import collections
import optparse
import os
import re
import sys
import thread
from datetime import datetime
from ..core.calculation import CalculationSettings
from ..core import file
from ..config.configuration import config


shell_colors = {
    "white_font": "\033[37m",
    "red_font": "\033[31m",
    "yellow_font": "\033[33m",
    "bold_font": "\033[1m",
    "red_background": "\033[41m",
    "default": "\033[0m",
}


get_progress_string = (
    lambda progress: "["
    + "=" * int(20 * progress)
    + " " * (20 - int(20 * progress))
    + "]"
    + " %.2f" % (100 * progress)
    + " %"
)


class ReadState(object):
    DEFAULT = 0
    ATOM_RADII = 1


class FileList(object):
    def __init__(self, global_resolution, global_atom_radius, global_export_dir):
        self.global_resolution = global_resolution
        self.global_atom_radius = global_atom_radius
        self.global_export_dir = global_export_dir
        self.file_list = []

    def append(
        self, filename, frames=None, resolution=None, atom_radii=None, export_dir=None
    ):
        self.file_list.append((filename, frames, resolution, atom_radii, export_dir))

    def createCalculationSettings(self, default_settings):
        settings_list = []
        for filename, frames, resolution, atom_radii, export_dir in self.file_list:
            if frames is None:
                frames = [-1]
            if resolution is None or self.global_resolution is not None:
                resolution = self.global_resolution
            if atom_radii is None or self.global_atom_radius is not None:
                atom_radii = self.global_atom_radius
            if export_dir is None or self.global_export_dir is not None:
                export_dir = self.global_export_dir
            settings = default_settings.copy()
            # TODO: atom radius
            settings.datasets = {filename: frames}
            settings.resolution = resolution
            settings.cutoff_radii = atom_radii
            settings.exportdir = export_dir
            settings_list.append(settings)
        return settings_list

    def __len__(self):
        return len(self.file_list)

    def __str__(self):
        s = ""
        for filename, frames, resolution, atom_radii, export_dir in self.file_list:
            s += "-> {}".format(filename)
            if frames is not None and frames[0] != -1:
                s += "; frames " + ", ".join([str(f + 1) for f in frames])
            else:
                s += "; all frames"
            if self.global_resolution is not None:
                s += "; resolution {}".format(self.global_resolution)
            elif resolution is not None:
                s += "; resolution {}".format(resolution)
            if self.global_atom_radius is not None:
                s += "; cutoff radius {}".format(self.global_atom_radius)
            elif atom_radii is not None:
                if isinstance(atom_radii, collections.abc.Mapping):
                    s += "; cutoff radii: {}".format(atom_radii)
                else:
                    s += "; cutoff radius {}".format(atom_radii)
            s += "\n"
        return s


class Cli(object):
    def __init__(self, control, command_line_params=None):
        if command_line_params is None:
            command_line_params = sys.argv[1:]

        self.control = control
        self.options_list = [
            {
                "special_type": None,
                "name": "quiet",
                "short": "-q",
                "long": "--quiet",
                "action": "store_true",
                "dest": "quiet",
                "default": False,
                "type": None,
                "help": "Less progress information is printed",
            },
            {
                "special_type": "parameter",
                "name": "atom radius",
                "short": "-a",
                "long": "--atomradius",
                "action": "store",
                "dest": "atom_radii",
                "default": None,
                "type": "float",
                "help": "cut off radius",
            },
            {
                "special_type": "parameter",
                "name": "resolution",
                "short": "-r",
                "long": "--resolution",
                "action": "store",
                "dest": "resolution",
                "default": 128,
                "type": "int",
                "help": "vacancy resolution",
            },
            {
                "special_type": "parameter",
                "name": "output directory",
                "short": "-o",
                "long": "--output_directory",
                "action": "store",
                "dest": "output_directory",
                "default": None,
                "type": "str",
                "help": "directory to store the results in",
            },
            {
                "special_type": "parameter",
                "name": "max cachefiles",
                "short": "",
                "long": "--maxcachefiles",
                "action": "store",
                "dest": "max_cachefiles",
                "default": None,
                "type": "int",
                "help": "maximum number of cached files",
            },
            {
                "special_type": None,
                "name": "no cache files",
                "short": "",
                "long": "--no-cache",
                "action": "store_true",
                "dest": "no_cache",
                "default": False,
                "type": None,
                "help": "Do not use any unnecessary cache files. One temporary cache file is always necessary.",
            },
            #                             {"special_type": "parameter",
            #                              "name": "bond delta",
            #                              "short": "-b",
            #                              "long": "--bonddelta",
            #                              "action": "store",
            #                              "dest": "bond_delta",
            #                              "default": None,
            #                              "type": "float",
            #                              "help": "constant bond delta (when no parameter is given, radii sum criterion (1.15) is used)"},
            {
                "special_type": "disable_target",
                "name": "no surface based cavities",
                "short": "",
                "long": "--nosurfacebasedcavities",
                "action": "store_true",
                "dest": "no_surface_based_cavities",
                "default": False,
                "type": None,
                "help": "Surface based cavities are not calculated.",
            },
            {
                "special_type": "disable_target",
                "name": "no center based cavities",
                "short": "",
                "long": "--nocenterbasedcavities",
                "action": "store_true",
                "dest": "no_center_based_cavities",
                "default": False,
                "type": None,
                "help": "Center based cavities are not calculated.",
            },
            {
                "special_type": "disable_target",
                "name": "no hdf5 export",
                "short": "",
                "long": "--nohdf5export",
                "action": "store_true",
                "dest": "no_hdf5_export",
                "default": False,
                "type": None,
                "help": "No HDF5 files with results are created",
            },
            {
                "special_type": "disable_target",
                "name": "no text export",
                "short": "",
                "long": "--notextexport",
                "action": "store_true",
                "dest": "no_text_export",
                "default": False,
                "type": None,
                "help": "No text files with results are created",
            },
        ]

        (self.options, self.left_args) = self.__parse_options(command_line_params)
        self.previous_progress = None
        self.cancel_callback = lambda: None

    # -------------------- public methods --------------------
    def start(self):
        file_list = self.__get_file_list(
            filter(lambda entry: entry[0] != "-", self.left_args)
        )

        default_settings = CalculationSettings(dict())
        default_settings.resolution = self.control.config.Computation.std_resolution
        default_settings.domains = True
        default_settings.surface_cavities = not self.options.no_surface_based_cavities
        default_settings.center_cavities = not self.options.no_center_based_cavities
        default_settings.recalculate = True
        default_settings.exporthdf5 = not self.options.no_hdf5_export
        default_settings.exporttext = not self.options.no_text_export
        default_settings.exportdir = None
        default_settings.bonds = True
        default_settings.dihedral_angles = True
        settings_list = file_list.createCalculationSettings(default_settings)
        if self.options.atom_radii is not None:
            config.Computation.atom_radii = self.options.atom_radii
        if self.options.max_cachefiles is not None:
            config.Computation.max_cachefiles = self.options.max_cachefiles
        elif self.options.no_cache:
            config.Computation.max_cachefiles = 1

        print("processing files:")
        print(file_list)
        print("parameters:")
        tmp_param_string = ""
        for i, opt_dict in enumerate(self.options_list):
            tmp_param_string += (
                opt_dict["name"]
                + ": "
                + str(getattr(self.options, opt_dict["dest"]))
                + ", "
                + ("\n" if i % 4 == 3 else "")
            )
        if tmp_param_string[-1] == "\n":
            tmp_param_string = tmp_param_string[:-3]
        else:
            tmp_param_string = tmp_param_string[:-2]
        print(tmp_param_string)
        print()
        print("=" * 80)

        if len(settings_list) > 0:
            print("Started calculation: {}".format(datetime.now()))
            started_computation = False
            for settings in settings_list:
                print("Calculating File:")
                for filename, _ in settings.datasets.items():
                    print(filename)
                print("...")
                sys.stdout.flush()
                results = self.control.calculation.calculate(settings)
                print("Done.")
            print("Finished calculation: {}".format(datetime.now()))

            #    started_computation = True

            # if started_computation:
            #    #self.__handle_input()
            #    try:
            #        self.control.request_action(Actions.Computation.WAIT_UNTIL_FINISHED, [True])
            #    except KeyboardInterrupt:
            #        try:
            #            self.cancel_callback()
            #        except task.IllegalTaskStateException:
            #            pass

    # -------------------- private methods -------------------

    def __handle_input(self):
        while True:
            try:
                line = raw_input()
            except EOFError:
                self.cancel_callback()
                break
            except KeyboardInterrupt:
                self.cancel_callback()
                break
            except ValueError:  # Wird beim Schliessen von stdin geworfen
                self.cancel_callback()
                break
        print()

    def __parse_options(self, command_line_params):
        usage = """Usage: %prog [options] batch_file1 batch_file2 ...

Batch files have the format:

RESOLUTION resolution
ATOM_RADIUS radius  /  ATOM_RADII (followed by indented lines with format "ELEMENT RADIUS")
OUTPUT_DIRECTORY directory
file-01.xyz
file-02.xyz frame frame ...
    .
    .
    .

RESOLUTION, ATOM_RADIUS and OUTPUT_DIRECTORY are optional and are overridden by the command line options. OUTPUT_DIRECTORY can contain one or more asterisks. These are replaced with the subdirectory name(s) containing the input files.
Note: Because the cutoff radius is stored in the global configuration, it cannot yet be specified per calculation."""
        parser = optparse.OptionParser(usage=usage)

        for opt_dict in self.options_list:
            parser.add_option(
                opt_dict["short"],
                opt_dict["long"],
                action=opt_dict["action"],
                dest=opt_dict["dest"],
                default=opt_dict["default"],
                type=opt_dict["type"],
                help=opt_dict["help"],
            )

        (options, left_args) = parser.parse_args(command_line_params)
        if len(left_args) == 0:
            parser.print_help()
            sys.exit()
        return (options, left_args)

    def __get_file_list(self, input_file_list):
        result_file_list = FileList(
            self.options.resolution,
            self.options.atom_radii,
            self.options.output_directory,
        )
        for input_file in input_file_list:
            read_state = ReadState.DEFAULT
            try:
                input_file = core.file.get_abspath(input_file)
                with open(input_file) as f:
                    resolution = None
                    atom_radii = None
                    output_directory = None
                    for line in f.readlines():
                        lstripped_line = line.lstrip()
                        if len(lstripped_line) > 0 and lstripped_line[0] != "#":
                            line_parts = lstripped_line.split()
                            line_is_processed = False
                            while not line_is_processed:
                                if read_state == ReadState.ATOM_RADII:
                                    if re.match(r"^\s+", line):
                                        element = line_parts[0]
                                        radius = float(line_parts[1])
                                        atom_radii[element] = (
                                            radius  # TODO ask ingo what is happening here
                                        )
                                        line_is_processed = True
                                    else:
                                        read_state = ReadState.DEFAULT
                                else:
                                    if line_parts[0] == "RESOLUTION":
                                        resolution = int(line_parts[1])
                                    elif line_parts[0] == "ATOM_RADIUS":
                                        atom_radii = float(line_parts[1])
                                    elif line_parts[0] == "ATOM_RADII":
                                        atom_radii = {}
                                        read_state = ReadState.ATOM_RADII
                                    elif line_parts[0] == "OUTPUT_DIRECTORY":
                                        output_directory = " ".join(line_parts[1:])
                                    else:
                                        filepath = os.path.join(
                                            os.path.dirname(
                                                os.path.abspath(input_file)
                                            ),
                                            line_parts[0],
                                        )
                                        if len(line_parts) > 1:
                                            frames = [
                                                int(f) - 1 for f in line_parts[1:]
                                            ]
                                        else:
                                            frames = None
                                        result_file_list.append(
                                            filepath,
                                            frames,
                                            resolution,
                                            atom_radii,
                                            output_directory,
                                        )
                                    line_is_processed = True
            except IOError:
                print(
                    "warning: batch file %s not accessable and skipped"
                    % (os.path.abspath(input_file))
                )

        return result_file_list



# -*- coding: utf-8 -*-


from __future__ import absolute_import


from PySide6 import QtCore, QtWidgets
from .util.table_with_removeable_items import TableWithRemoveableEntries
from .util.cutoff_preview_table import CutoffPreviewTable


class CutoffPresetTable(TableWithRemoveableEntries):
    HEADER_LABELS = ("Preset Name", "Cutoff")

    def __init__(self, presets):
        self._presets = list(presets)
        super().__init__(self.HEADER_LABELS, len(presets))
        self._orig_index_to_row_index = range(len(self._presets))
        self._indices_to_remove = set()
        self._init_ui()

    def _init_ui(self):
        pass

    @property
    def presets(self):
        return [
            entry
            for i, entry in enumerate(self._presets)
            if i in self._indices_to_remove
        ]

    @property
    def preset_entries_for_deletion(self):
        return [self._presets[i] for i in sorted(self._indices_to_remove)]

    def entry_count(self):
        return len(self._presets)

    def entries(self):
        entries = []
        for i, presets_entry in enumerate(self._presets):
            current_row = []
            for j, value in enumerate((presets_entry.name,)):
                item = QtWidgets.QTableWidgetItem(str(value))
                item.setFlags(item.flags() ^ QtCore.Qt.ItemIsEditable)
                current_row.append(item)
            if isinstance(presets_entry.radii, float):
                cutoff_fixed_value = presets_entry.radii
                item = QtWidgets.QTableWidgetItem(str(cutoff_fixed_value))
                item.setFlags(item.flags() ^ QtCore.Qt.ItemIsEditable)
                item.setTextAlignment(QtCore.Qt.AlignCenter)
                current_row.append(item)
            else:
                cutoff_table = CutoffPreviewTable(presets_entry.radii)
                current_row.append(cutoff_table)
            entries.append(tuple(current_row))
        return entries



# -*- coding: utf-8 -*-

from PySide6 import QtWidgets
from .gl_widget import GLWidget
from .pdf_widget import PDFWidget
from .histogram_widget import HistogramWidget


class GLStack(QtWidgets.QStackedWidget):
    def __init__(self, parent, main_window):
        QtWidgets.QStackedWidget.__init__(self, parent)
        self.parent = parent
        self.control = parent.control
        self.gl_widget = GLWidget(self, main_window)
        self.pdf_widget = PDFWidget(self)
        self.histogram_widget = HistogramWidget(self)

        self.addWidget(self.gl_widget)
        self.addWidget(self.pdf_widget)
        self.addWidget(self.histogram_widget)

        self.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        )

        self.show()

    def activate(self, index):
        self.setCurrentIndex(index)
        self.widget(index).activate()

    def updatestatus(self):
        for index in range(self.count()):
            widget = self.widget(index)
            widget.updatestatus()



# -*- coding: utf-8 -*-
"""
This module provides classes to handle pyMolDyn related files.
"""

import os
import os.path
import h5py
from itertools import repeat
from datetime import datetime
from . import data
import sys
from ..util.logger import Logger
from . import elements

try:
    import pybel

    raise ImportError
except ImportError:

    class pybel:
        """
        Dummy class representing a missing pybel module.
        """

        informats = {}


__all__ = ["File", "InputFile", "ResultFile", "XYZFile", "HDF5File"]

logger = Logger("core.file")
logger.setstream("default", sys.stdout, Logger.WARNING)

SEARCH_PATH = None


def get_abspath(path):
    """
    Return an absolute path for an input file. The difference to abspath from
    os.path is that this function uses SEARCH_PATH instead of the current
    working directory
    """
    global SEARCH_PATH
    if not os.path.isabs(path) and SEARCH_PATH is not None:
        # relative paths use the SEARCH_PATH instead of the current working
        # directory, as that is modified at startup. The SEARCH_PATH is set
        # to the initial cwd in startBatch and startGUI.
        path = os.path.join(SEARCH_PATH, path)
    return os.path.abspath(path)


class FileError(Exception):
    """
    Exception which will be raised by the subclasses of :class:`InputFile`
    if errors occur when handling the data inside the file.
    """

    def __init__(self, message, cause=None):
        super().__init__(message, cause)
        self.message = message
        self.cause = cause

    def __str__(self):
        if self.cause is None:
            f = "{}: {}"
        else:
            f = "{}: {} (caused by: {})"
        return f.format(self.__class__.__name__, self.message, str(self.cause))


class InputFile(object):
    """
    Abstract access to a file that contains atom data for one or more frames.
    Subclasses need to implement the ``readinfo()`` and ``readatoms(frame)`` methods.
    """

    def __init__(self, path):
        """
        **Parameters:**
            `path` :
                absolute path to the file
        """
        self.path = get_abspath(path)
        self._info = data.FileInfo()
        self.inforead = False

    @property
    def info(self):
        """
        :class:`core.data.FileInfo` object that contains metadata
        """
        if not self.inforead:
            try:
                self.readinfo()
            except IOError as e:
                # logger.error(str(e))
                pass
            if self._info.volume is None:
                self._info.volume_guessed = True
                minx, maxx = float("inf"), float("-inf")
                miny, maxy = float("inf"), float("-inf")
                minz, maxz = float("inf"), float("-inf")
                for frame in range(self._info.num_frames):
                    atoms = self.getatoms(frame)
                    minx = min(minx, atoms.positions[:, 0].min())
                    maxx = max(maxx, atoms.positions[:, 0].max())
                    miny = min(miny, atoms.positions[:, 1].min())
                    maxy = max(maxy, atoms.positions[:, 1].max())
                    minz = min(minz, atoms.positions[:, 2].min())
                    maxz = max(maxz, atoms.positions[:, 2].max())
                self._info.volumestr = "ORT %f %f %f" % (
                    maxx - minx,
                    maxy - miny,
                    maxz - minz,
                )
        return self._info

    def getatoms(self, frame):
        """
        Read atom data for a specified frame.

        **Parameters:**
            `frame` :
                the frame number

        **Returns:**
            an :class:`core.data.Atoms` object

        **Raises:**
            - :class:`IndexError`: if the frame is not in the file
            - :class:`FileError`: if there are problems with the data in the file
            - :class:`IOError`: if the file cannot be read
        """
        return self.readatoms(frame)

    def readinfo(self):
        raise NotImplementedError

    def readatoms(self, frame):
        raise NotImplementedError


class XYZFile(InputFile):
    """
    Implementation on :class:`InputFile` for Open Babel 'xyz' files.
    """

    def __init__(self, path):
        super().__init__(path)
        f = open(self.path, "r")
        f.close()

    def readinfo(self):
        try:
            self._info.num_frames = 0
            with open(self.path.encode("utf-8"), "r") as f:
                try:
                    while True:
                        num_atoms = f.readline()
                        if not num_atoms:
                            break
                        num_atoms = int(num_atoms)
                        volume_info = f.readline()
                        for i in range(num_atoms):
                            f.readline()
                        self._info.num_frames += 1
                        if self._info.num_frames == 1:
                            self._info.volumestr = volume_info
                except StopIteration:
                    pass
            self.inforead = True
        except IOError:
            raise
        except Exception as e:
            raise FileError("Cannot read file info.", e)

    def readatoms(self, frame):
        try:
            if self.info.num_frames <= frame:
                raise IndexError("Frame {} not found".format(frame))
            with open(self.path.encode("utf-8"), "r") as f:
                try:
                    # Skip the first frames
                    for i in range(frame):
                        num_atoms = f.readline()
                        if not num_atoms:
                            break
                        num_atoms = int(num_atoms)
                        f.readline()
                        for i in range(num_atoms):
                            f.readline()
                    # actually read the molecule
                    symbols = []
                    positions = []
                    num_atoms = f.readline()
                    if not num_atoms:
                        raise StopIteration
                    num_atoms = int(num_atoms)
                    f.readline()
                    for i in range(num_atoms):
                        line = f.readline()
                        if line.strip():
                            symbol, x, y, z = line.split()[:4]
                            position = (float(x), float(y), float(z))
                            symbols.append(symbol)
                            positions.append(position)
                except StopIteration:
                    raise IndexError("Frame {} not found".format(frame))
            return data.Atoms(positions, None, symbols, self.info.volume)
        except (IOError, IndexError):
            raise
        except Exception as e:
            raise FileError("Cannot read atom data.", e)


class BabelFile(InputFile):
    """
    Implementation on :class:`InputFile` for Open Babel 'xyz' files.
    """

    def __init__(self, path):
        super().__init__(path)
        # Check if the file exists
        f = open(self.path, "r")
        f.close()

    def readinfo(self):
        try:
            file_extension = os.path.splitext(self.path)[1][1:]
            mol_iter = pybel.readfile(
                file_extension.encode("utf8"), self.path.encode("utf8")
            )
            try:
                mol = next(mol_iter)
                self._info.volumestr = mol.title
                self._info.num_frames = 1
                for _ in mol_iter:
                    self._info.num_frames += 1
            except StopIteration:
                self._info.num_frames = 0
            self.inforead = True
        except IOError:
            raise
        except Exception as e:
            raise FileError("Cannot read file info.", e)

    def readatoms(self, frame):
        try:
            if self.info.num_frames <= frame:
                raise IndexError("Frame {} not found".format(frame))

            file_extension = os.path.splitext(self.path)[1][1:]
            mol_iter = pybel.readfile(
                file_extension.encode("utf8"), self.path.encode("utf8")
            )

            # get the correct frame
            try:
                for _ in range(frame):
                    mol_iter.next()
                mol = mol_iter.next()
            except StopIteration:
                raise IndexError("Frame {} not found".format(frame))

            # read the atom information
            symbols = []
            positions = []
            for atom in mol.atoms:
                positions.append(tuple(float(c) for c in atom.coords))
                symbol = core.elements.symbols[atom.atomicnum]
                symbols.append(symbol)
            return data.Atoms(positions, None, symbols, self.info.volume)
        except (IOError, IndexError):
            raise
        except Exception as e:
            raise FileError("Cannot read atom data.", e)


class ResultFile(InputFile):
    """
    Abstract access to a file that contains both input data (atoms) and
    calculated results.
    The `info` attribute has the type `ResultInfo`.
    In addition the `readinfo` and `readatoms` methods, subclasses must
    implement `writeinfo`, `readresults` and `writeresults`.
    """

    def __init__(self, path, sourcefilepath=None):
        """
        **Parameters:**
            `path` :
                absolute path to the file

            `sourcefilepath` :
                path to the file where the input data originally came from
        """
        super().__init__(path)
        self._info = data.ResultInfo()
        self._info.sourcefilepath = sourcefilepath

    def getresults(self, frame, resolution):
        """
        Read results from this file.

        **Parameters:**
            `frame` :
                the frame number
            `resolution` :
                the resolution of the calculation

        **Returns:**
            A `Results` object, if the file contains results for the
            specified parameters. Otherwise it return `None`.

        **Raises:**
            - :class:`FileError`: if there are problems with the data in the file
            - :class:`IOError`: if the file cannot be read
        """
        if not self.info[resolution].domains[frame] is None:
            return self.readresults(frame, resolution)
        else:
            return None

    def addresults(self, results, overwrite=True):
        """
        Write calculated results into this file.

        **Parameters:**
            `results` :
                the results to write
            `overwrite` :
                specifies if existing results should be overwritten

        **Raises:**
            - :class:`FileError`: if there are problems with the data in the file
            - :class:`IOError`: if the file cannot be read or written
        """
        self.writeresults(results, overwrite=overwrite)
        resinfo = self.info[results.resolution]
        if results.domains:
            resinfo.domains[results.frame] = results.domains.timestamp
        if results.surface_cavities:
            resinfo.surface_cavities[results.frame] = results.surface_cavities.timestamp
        if results.center_cavities:
            resinfo.center_cavities[results.frame] = results.center_cavities.timestamp
        self.writeinfo()

    def writeinfo(self):
        raise NotImplementedError

    def readresults(self, frame, resolution):
        raise NotImplementedError

    def writeresults(self, results, overwrite=True):
        raise NotImplementedError


class HDF5File(ResultFile):
    """
    Implementation on :class:`ResultFile` for 'hdf5' files.
    """

    def __init__(self, path, sourcefilepath=None):
        super().__init__(path, sourcefilepath)

    @classmethod
    def fromInputFile(cls, filepath, sourcefilepath):
        """
        Create a new :class:`HDF5File`. Copy atoms from an
        :class:`InputFile`. This is used to initially create a file
        to export results to.
        """
        inputfile = File.open(sourcefilepath)
        outputfile = cls(filepath, sourcefilepath)
        for frame in range(inputfile.info.num_frames):
            atoms = inputfile.getatoms(frame)
            results = data.Results(filepath, frame, 64, atoms, None, None, None)
            outputfile.writeresults(results)
        outputfile.readinfo()
        outputfile.writeinfo()
        return outputfile

    def readatoms(self, frame):
        atoms = None
        if not os.path.isfile(self.path):
            raise IOError(2, "File not found.")
        try:
            with h5py.File(self.path) as f:
                group = "atoms/frame{}".format(frame)
                if group not in f:
                    raise IndexError("Frame {} not found".format(frame))
                atoms = data.Atoms(f[group])
        except (IOError, IndexError):
            raise
        except Exception as e:
            raise FileError("Cannot read atom data.", e)
        return atoms

    def readinfo(self):
        if not os.path.isfile(self.path):
            raise IOError(2, "File not found.")
        try:
            with h5py.File(self.path) as f:
                if "info" in f:
                    info = data.ResultInfo(f["info"])
                    if self._info.sourcefilepath is not None:
                        info.sourcefilepath = self._info.sourcefilepath
                    self._info = info
                    self.inforead = True
        except IOError:
            raise
        except Exception as e:
            raise FileError("Cannot read file info.", e)

        if (
            not self.inforead
            and self._info.sourcefilepath is not None
            and os.path.isfile(self._info.sourcefilepath)
        ):
            try:
                sf = File.open(self._info.sourcefilepath)
                self._info.num_frames = sf.info.num_frames
                self._info.volumestr = sf.info.volumestr
                self.inforead = True
            except IOError:
                raise
            except Exception as e:
                raise FileError("Cannot read file info.", e)

        if not self.inforead:
            raise RuntimeError("No File Info in this file and the source file.")

    def writeinfo(self):
        try:
            with h5py.File(self.path, "a") as f:
                h5group = f.require_group("info")
                self.info.tohdf(h5group)
        except IOError:
            raise
        except Exception as e:
            raise FileError("Cannot write file info.", e)

    def readresults(self, frame, resolution):
        if not os.path.isfile(self.path):
            raise IOError(2, "File not found.")
        try:
            results = None
            with h5py.File(self.path) as f:
                groupname = "results/frame{}/resolution{}".format(frame, resolution)
                if groupname in f:
                    group = f[groupname]
                    atoms = data.Atoms(f["atoms/frame{}".format(frame)])
                    domains = data.Domains(group["domains"])
                    if "surface_cavities" in group:
                        surface_cavities = data.Cavities(group["surface_cavities"])
                    else:
                        surface_cavities = None
                    if "center_cavities" in group:
                        center_cavities = data.Cavities(group["center_cavities"])
                    else:
                        center_cavities = None
                    if self.info.sourcefilepath is not None:
                        filepath = self.info.sourcefilepath
                    else:
                        filepath = self.path
                    results = data.Results(
                        filepath,
                        frame,
                        resolution,
                        atoms,
                        domains,
                        surface_cavities,
                        center_cavities,
                    )
        except IOError:
            raise
        except Exception as e:
            raise FileError("Cannot read results.", e)
        return results

    def writeresults(self, results, overwrite=True):
        # TODO: results valid?
        try:
            with h5py.File(self.path, "a") as f:
                group = f.require_group("atoms/frame{}".format(results.frame))
                # TODO: is it OK to never overwrite atoms?
                results.atoms.tohdf(group, overwrite=False)
                if (
                    results.domains is not None
                    or results.surface_cavities is not None
                    or results.center_cavities is not None
                ):
                    group = f.require_group(
                        "results/frame{}/resolution{}".format(
                            results.frame, results.resolution
                        )
                    )
                if results.domains is not None:
                    subgroup = group.require_group("domains")
                    results.domains.tohdf(subgroup, overwrite=overwrite)
                if results.surface_cavities is not None:
                    subgroup = group.require_group("surface_cavities")
                    results.surface_cavities.tohdf(subgroup, overwrite=overwrite)
                if results.center_cavities is not None:
                    subgroup = group.require_group("center_cavities")
                    results.center_cavities.tohdf(subgroup, overwrite=overwrite)
        except IOError:
            raise
        except Exception as e:
            raise FileError("Cannot write results.", e)


class File(object):
    """
    Provides static methods for easy access to files and directories.
    The class attribute `types` associates filename endings with
    classes to handle them.
    """

    types = dict(
        list(zip(pybel.informats.keys(), repeat(BabelFile)))
        + [
            ("xyz", XYZFile),
            ("hdf5", HDF5File),
        ]
    )

    @classmethod
    def listdir(cls, directory):
        """
        List all (possible) pyMolDyn files in the directory.

        **Parameters:**
            `directory` :
                path to a directory

        **Returns:**
            A list of filenames which can be opened.
        """
        if not directory:
            directory = "."
        return [f for f in os.listdir(directory) if cls.exists(f)]

    @classmethod
    def open(cls, filepath):
        """
        Get the associated :class:`InputFile` object for the given file.

        **Parameters:**
            `filepath` :
                path to the file

        **Returns:**
            An object of a subclass of :class:`InputFile`.

        **Raises:**
            :class:`ValueError` if the file format is unknown.
        """
        e = filepath.split(".")[-1]
        if e not in cls.types:
            raise ValueError("Unknown file format")
        FileClass = cls.types[e]
        return FileClass(filepath)

    @classmethod
    def exists(cls, filepath):
        """
        Check if a file exists and if it can be opened.

        **Parameters:**
            `filepath` :
                path to the file

        **Returns:**
            `True` if the file exists and there is a subclass of :class:`InputFile`
            associated with the filename ending.
        """
        filepath = get_abspath(filepath)
        name = os.path.basename(filepath)
        return os.path.isfile(filepath) and name.split(".")[-1] in cls.types



# -*- coding: utf-8 -*-


from __future__ import absolute_import


import collections
from PySide6 import QtCore, QtWidgets
from .table_fit import TableFit


class CutoffPreviewTable(QtWidgets.QTableWidget):
    def __init__(self, radii):
        super().__init__(1, len(radii))
        self._table_fit = TableFit(self, scrollbar_extra_space=(0, 0))
        self._radii = collections.OrderedDict(
            sorted(radii.items())
        )  # Sorted by element name
        self._init_ui()

    def _init_ui(self):
        self.setHorizontalHeaderLabels(self._radii.keys())
        self.setVerticalHeaderLabels(("  ",))
        self.setShowGrid(True)
        self.setSelectionMode(QtWidgets.QAbstractItemView.NoSelection)
        for i, cutoff_radius in enumerate(self._radii.values()):
            item = QtWidgets.QTableWidgetItem(str(cutoff_radius))
            item.setFlags(item.flags() ^ QtCore.Qt.ItemIsEditable)
            self.setItem(0, i, item)



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import collections
import json
import os.path
import sys
from .configuration import CONFIG_DIRECTORY
from ..util.logger import Logger

logger = Logger("config.cutoff_presets")
logger.setstream("default", sys.stdout, Logger.WARNING)


DEFAULT_CONFIG_FILE = os.path.expanduser(
    os.path.join(CONFIG_DIRECTORY, "cutoff_presets.json")
)


Preset = collections.namedtuple("Preset", ["name", "radii"])


class CutoffPresets(object):
    def __init__(self, config_filepath=DEFAULT_CONFIG_FILE):
        self._config_filepath = config_filepath
        self._presets = None
        self.read()

    def read(self):
        try:
            with open(self._config_filepath, "r") as f:
                presets_json = json.load(f)
            self._presets = [Preset(*entry) for entry in presets_json]
        except IOError:
            self._presets = []

    def save(self):
        try:
            with open(self._config_filepath, "w") as f:
                json.dump(self._presets, f)
        except IOError:
            logger.warn("Could not save cutoff radii presets")
            raise

    @property
    def presets(self):
        return list(self._presets)

    def add(self, entry):
        self._presets.insert(0, entry)

    def extend(self, entry_list):
        for entry in entry_list:
            self.add(entry)

    def remove(self, entry):
        self._presets.remove(entry)

    def remove_list(self, entry_list):
        for entry in entry_list:
            self.remove(entry)


cutoff_presets = CutoffPresets()
